unit MCoPRotatableImage2;

{
        TRotatableImage
         Version 2.0.1
      By Mohammad Yousefi
    vahid_you2004@yahoo.com
             2013
   http://www.dihav.com/
                              }

interface

uses
  Windows, SysUtils, Messages, Types, Classes,
  {$if CompilerVersion > 22}
    Vcl.Graphics, Vcl.ExtCtrls, Vcl.Controls, Vcl.Forms
  {$else}
    Graphics, ExtCtrls, Controls, Forms
  {$ifend}
  ;

type

  TMappingEvent = procedure (Sender:TObject;AU,AV:Word;var AX,AY:Single) of object;

  TRotatableImage = class(TGraphicControl)
  private
    RA,RR,fRA:Real;
    Data,Rotated,Stretched,CustomMap:Pointer;
    fPicture:TPicture;
    fOnProgress:TProgressEvent;
    fStretch,fCenter,fIncrementalDisplay,fTransparent,fDrawing,fProportional,fCM:Boolean;
    fBlendColor:TColor;
    BlendCol:Cardinal;
    fCME:TMappingEvent;
    procedure SetRA(V:Real);
    procedure SetHome(V:String);
    function GetHome:String;
    function FindColor(RR,RG,RB,BR,BG,BB:Byte):Cardinal;
    function GetCanvas:TCanvas;
    procedure PictureChanged(Sender:TObject);
    procedure SetCenter(V:Boolean);
    procedure SetPicture(V:TPicture);
    procedure SetStretch(V:Boolean);
    procedure SetTransparent(V:Boolean);
    procedure SetProportional(V:Boolean);
    procedure MakeData;
    procedure MakeRotated;
    procedure MakeStretched;
    function RotatedWidth:Word;
    function RotatedHeight:Word;
    procedure ReadHandle(H:HIcon);
    function GetPixel(Memory:Pointer;X,Y,Width,Height:Integer):Cardinal;
    procedure SetBlendColor(V:TColor);
    function StretchedWidth:Word;
    function StretchedHeight:Word;
    procedure SetCM(V:Boolean);
  protected
    function CanAutoSize(var NewWidth,NewHeight:Integer):Boolean; override;
    function DoPaletteChange:Boolean;
    function GetPalette:HPALETTE; override;
    procedure Paint; override;
    procedure Progress(Sender:TObject;Stage:TProgressStage;PercentDone:Byte;RedrawNow:Boolean;const R:TRect;const Msg:String); dynamic;
    procedure Resize; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Loaded; override;
    property Canvas:TCanvas read GetCanvas;
    procedure MakeCustomMapping;
  published
    property Align;
    property Anchors;
    property AutoSize;
    property BlendColor:TColor read fBlendColor write SetBlendColor default clBtnFace;
    property Center:Boolean read fCenter write SetCenter default False;
    property Constraints;
    property CustomMapping:Boolean read fCM write SetCM default False;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Homepage:String read GetHome write SetHome stored False;
    property IncrementalDisplay:Boolean read fIncrementalDisplay write fIncrementalDisplay default False;
    property ParentShowHint;
    property Picture:TPicture read fPicture write SetPicture;
    property PopupMenu;
    property Proportional:Boolean read fProportional write SetProportional default False;
    property RotationAngle:Real read fRA write SetRA;
    property ShowHint;
    property Stretch:Boolean read fStretch write SetStretch default False;
    property Transparent:Boolean read FTransparent write SetTransparent default False;
    property Visible;
    property OnClick;
    property OnContextPopup;
    property OnCustomMapping:TMappingEvent read fCME write fCME;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnProgress:TProgressEvent read fOnProgress write fOnProgress;
    property OnStartDock;
    property OnStartDrag;
  end;

procedure Register;

implementation

constructor TRotatableImage.Create(AOwner: TComponent);
begin
inherited Create(AOwner);
ControlStyle:=ControlStyle+[csReplicatable];
fPicture:=TPicture.Create;
fPicture.OnChange:=PictureChanged;
fPicture.OnProgress:=Progress;
Height:=105;
Width:=105;
fRA:=0;
RA:=0;
RR:=0;
fCM:=False;
fOnProgress:=nil;
fCME:=nil;
SetBlendColor(clBtnFace);
Data:=nil;
Rotated:=nil;
Stretched:=nil;
CustomMap:=nil;
end;

destructor TRotatableImage.Destroy;
begin
FreeMemory(Data);
FreeMemory(Rotated);
FreeMemory(Stretched);
FreeMemory(CustomMap);
fPicture.Free;
inherited Destroy;
end;

procedure TRotatableImage.Loaded;
begin
inherited Loaded;
PictureChanged(Self);
end;

procedure TRotatableImage.SetBlendColor(V:TColor);
begin
if fBlendColor<>V then
  begin
  fBlendColor:=V;
  BlendCol:=ColorToRGB(V) and $00FFFFFF;
	PictureChanged(Self);
  end;
end;

function TRotatableImage.GetPalette:HPALETTE;
begin
Result:=0;
if fPicture.Graphic<>nil then Result:=fPicture.Graphic.Palette;
end;

function TRotatableImage.DoPaletteChange:Boolean;
var
  ParentForm:TCustomForm;
  Tmp:TGraphic;
begin
Result:=False;
Tmp:=Picture.Graphic;
if Visible and (not (csLoading in ComponentState)) and (Tmp<>nil) and (Tmp.PaletteModified) then
  begin
	if (Tmp.Palette=0) then
    Tmp.PaletteModified:=False
	else
	  begin
	  ParentForm:=GetParentForm(Self);
	  if Assigned(ParentForm) and ParentForm.Active and Parentform.HandleAllocated then
	    begin
		  if fDrawing then
		    ParentForm.Perform(wm_QueryNewPalette,0,0)
		  else
		    PostMessage(ParentForm.Handle,wm_QueryNewPalette,0,0);
		  Result:=True;
		  Tmp.PaletteModified:=False;
	    end;
	  end;
  end;
end;

procedure TRotatableImage.Progress(Sender:TObject;Stage:TProgressStage;PercentDone:Byte;RedrawNow:Boolean;const R:TRect;const Msg:String);
begin
if fIncrementalDisplay and RedrawNow then
  begin
	if DoPaletteChange then
    Update
	else
    Paint;
  end;
if Assigned(fOnProgress) then fOnProgress(Sender,Stage,PercentDone,RedrawNow,R,Msg);
end;

function TRotatableImage.GetCanvas:TCanvas;
var Bitmap:TBitmap;
begin
if Picture.Graphic=nil then
  begin
	Bitmap:=TBitmap.Create;
	try
	  Bitmap.Width:=Width;
	  Bitmap.Height:=Height;
	  Picture.Graphic:=Bitmap;
	finally
	  Bitmap.Free;
	end;
  end;
if Picture.Graphic is TBitmap then
	Result:=TBitmap(Picture.Graphic).Canvas
else
	raise EInvalidOperation.Create('Can only modify an image if it contains a bitmap');
end;

procedure TRotatableImage.SetCenter(V:Boolean);
begin
if fCenter<>V then
  begin
	fCenter:=V;
	PictureChanged(Self);
  end;
end;

procedure TRotatableImage.SetPicture(V:TPicture);
begin
fPicture.Assign(V);
MakeData;
end;

procedure TRotatableImage.SetStretch(V:Boolean);
begin
if V<>fStretch then
  begin
	fStretch:=V;
	PictureChanged(Self);
  end;
end;

procedure TRotatableImage.SetTransparent(V:Boolean);
begin
if V<>fTransparent then
  begin
	fTransparent:=V;
	PictureChanged(Self);
  end;
end;

procedure TRotatableImage.SetProportional(V:Boolean);
begin
if fProportional<>V then
  begin
	fProportional:=V;
	PictureChanged(Self);
  end;
end;

procedure TRotatableImage.SetCM(V:Boolean);
begin
if fCM<>V then
  begin
	fCM:=V;
	PictureChanged(Self);
  end;
end;

procedure TRotatableImage.PictureChanged(Sender:TObject);
var G:TGraphic;
begin
if AutoSize and (RotatedWidth>0) and (RotatedHeight>0) and (not fCM) then
	SetBounds(Left,Top,RotatedWidth,RotatedHeight);
G:=Picture.Graphic;
if G<>nil then
  begin
	if not ((G is TMetaFile) or (G is TIcon)) then G.Transparent:=fTransparent;
	if DoPaletteChange and fDrawing then Update;
  end;
MakeData;
if not fDrawing then Invalidate;
end;

function TRotatableImage.CanAutoSize(var NewWidth,NewHeight:Integer):Boolean;
begin
Result:=True;
if not (csDesigning in ComponentState) or (RotatedWidth>0) and (RotatedHeight>0) then
  begin
  if Align in [alNone,alLeft,alRight] then  NewWidth:=RotatedWidth;
  if Align in [alNone,alTop,alBottom] then  NewHeight:=RotatedHeight;
  end;
end;

function TRotatableImage.GetHome:String;
begin
Result:='http://www.dihav.com/';
end;

procedure TRotatableImage.SetHome(V:String);
begin
//
end;

procedure TRotatableImage.SetRA(V:Real);
begin
if fRA<>V then
  begin
  fRA:=V;
  RA:=V;
  while RA>=360 do RA:=RA-360;
  while RA<0 do RA:=RA+360;
  RR:=RA*PI/180;
  PictureChanged(Self);
  end;
end;

procedure TRotatableImage.MakeCustomMapping;
var
  w,h,mw,mh,x,y:Word;
  c,u,a1,a2,a3,a4,ch,cw:Single;
  l1,l2,l3,l4:Cardinal;
  d:PCardinal;
  tc,tu:Integer;

  function MakeColor:Cardinal;
  var r,g,b,a:Byte;
  begin
  r:=Round(Byte(l1)*a1+Byte(l2)*a2+Byte(l3)*a3+Byte(l4)*a4);
  g:=Round(Byte(l1 shr 8)*a1+Byte(l2 shr 8)*a2+Byte(l3 shr 8)*a3+Byte(l4 shr 8)*a4);
  b:=Round(Byte(l1 shr 16)*a1+Byte(l2 shr 16)*a2+Byte(l3 shr 16)*a3+Byte(l4 shr 16)*a4);
  a:=Round(Byte(l1 shr 24)*a1+Byte(l2 shr 24)*a2+Byte(l3 shr 24)*a3+Byte(l4 shr 24)*a4);
  Result:=r or (g shl 8) or (b shl 16) or (a shl 24);
  end;

  function Tru(V:Single):Integer;
  begin
  if V<0 then Result:=-1 else Result:=Trunc(V);
  end;

begin
FreeMemory(CustomMap);
CustomMap:=nil;

if Data=nil then Exit;

if not fCM then Exit;
if not Assigned(fCME) then Exit;

if Picture.Graphic=nil then Exit;
if Picture.Graphic is TIcon then ReadHandle(TIcon(Picture.Graphic).Handle);
mw:=Picture.Width;
mh:=Picture.Height;
if (mw=0) or (mh=0) then Exit;
w:=ClientWidth;
h:=ClientHeight;

CustomMap:=GetMemory(w*h*4);
FillMemory(CustomMap,w*h*4,0);
d:=CustomMap;

cw:=w/mw;
ch:=h/mh;

for y:=h-1 downto 0 do
  for x:=0 to w-1 do
    begin
    c:=x/cw;
    u:=y/ch;
    fCME(Self,x,y,c,u);
    if (c<=-1) or (u<=-1) or (c>=mw) or (u>=mh) then
      begin
      Inc(d);
      Continue;
      end;
    tc:=Tru(c);
    tu:=Tru(u);
    if (c<0) or (u<0) then l1:=BlendCol else l1:=GetPixel(Data,tc,tu,mw,mh);
    if (tc=mw-1) or (u<0) then l2:=BlendCol else l2:=GetPixel(Data,tc+1,tu,mw,mh);
    if (c<0) or (tu=mh-1) then l3:=BlendCol else l3:=GetPixel(Data,tc,tu+1,mw,mh);
    if (tc=mw-1) or (tu=mh-1) then l4:=BlendCol else l4:=GetPixel(Data,tc+1,tu+1,mw,mh);
    c:=c-tc;
    u:=u-tu;
    a1:=(1-c)*(1-u);
    a2:=c*(1-u);
    a3:=(1-c)*u;
    a4:=c*u;
    d^:=MakeColor;
    Inc(d);
    end;
end;

procedure TRotatableImage.MakeStretched;
var
  w,h,x,y,rw,rh,tc,tu:Word;
  cw,ch,c,u,a1,a2,a3,a4:Single;
  d:PCardinal;
  from:Pointer;
  l1,l2,l3,l4:Cardinal;

  function MakeColor:Cardinal;
  var r,g,b,a:Byte;
  begin
  r:=Round(Byte(l1)*a1+Byte(l2)*a2+Byte(l3)*a3+Byte(l4)*a4);
  g:=Round(Byte(l1 shr 8)*a1+Byte(l2 shr 8)*a2+Byte(l3 shr 8)*a3+Byte(l4 shr 8)*a4);
  b:=Round(Byte(l1 shr 16)*a1+Byte(l2 shr 16)*a2+Byte(l3 shr 16)*a3+Byte(l4 shr 16)*a4);
  a:=Round(Byte(l1 shr 24)*a1+Byte(l2 shr 24)*a2+Byte(l3 shr 24)*a3+Byte(l4 shr 24)*a4);
  Result:=r or (g shl 8) or (b shl 16) or (a shl 24);
  end;

begin
FreeMemory(Stretched);
Stretched:=nil;

if not (Stretch or (Proportional and ((RotatedWidth>ClientWidth) or (RotatedHeight>ClientHeight)))) then Exit;
if Picture.Graphic=nil then Exit;
if (RotatedWidth=0) or (RotatedHeight=0) then Exit;

from:=Rotated;
if (Rotated=nil) and (RA<>0) then Exit;
if (Rotated=nil) and (RA=0) then from:=Data;
if from=nil then Exit;

rw:=RotatedWidth;
rh:=RotatedHeight;
w:=StretchedWidth;
h:=StretchedHeight;
cw:=w/rw;
ch:=h/rh;

Stretched:=GetMemory(w*h*4);
FillMemory(Stretched,w*h*4,0);
d:=Stretched;

for y:=h-1 downto 0 do
  for x:=0 to w-1 do
    begin
    c:=x/cw;
    u:=y/ch;
    tc:=Trunc(c);
    tu:=Trunc(u);
    l1:=GetPixel(from,tc,tu,rw,rh);
    if tc=rw-1 then l2:=BlendCol else l2:=GetPixel(from,tc+1,tu,rw,rh);
    if tu=rh-1 then l3:=BlendCol else l3:=GetPixel(from,tc,tu+1,rw,rh);
    if (tc=rw-1) or (tu=rh-1) then l4:=BlendCol else l4:=GetPixel(from,tc+1,tu+1,rw,rh);
    c:=c-tc;
    u:=u-tu;
    a1:=(1-c)*(1-u);
    a2:=c*(1-u);
    a3:=(1-c)*u;
    a4:=c*u;
    d^:=MakeColor;
    Inc(d);
    end;
end;

procedure TRotatableImage.MakeRotated;
var
  w,h,mw,mh,x,y:Word;
  c,u,a1,a2,a3,a4,si,co,mw2,mh2,w2,h2:Single;
  l1,l2,l3,l4:Cardinal;
  d:PCardinal;
  tc,tu:Integer;

  function MakeColor:Cardinal;
  var r,g,b,a:Byte;
  begin
  r:=Round(Byte(l1)*a1+Byte(l2)*a2+Byte(l3)*a3+Byte(l4)*a4);
  g:=Round(Byte(l1 shr 8)*a1+Byte(l2 shr 8)*a2+Byte(l3 shr 8)*a3+Byte(l4 shr 8)*a4);
  b:=Round(Byte(l1 shr 16)*a1+Byte(l2 shr 16)*a2+Byte(l3 shr 16)*a3+Byte(l4 shr 16)*a4);
  a:=Round(Byte(l1 shr 24)*a1+Byte(l2 shr 24)*a2+Byte(l3 shr 24)*a3+Byte(l4 shr 24)*a4);
  Result:=r or (g shl 8) or (b shl 16) or (a shl 24);
  end;

  function Tru(V:Single):Integer;
  begin
  if V<0 then Result:=-1 else Result:=Trunc(V);
  end;

begin
FreeMemory(Rotated);
Rotated:=nil;
FreeMemory(Stretched);
Stretched:=nil;

if Data=nil then Exit;

if RA=0 then
  begin
  MakeStretched;
  Exit;
  end;
if Picture.Graphic=nil then Exit;
w:=RotatedWidth;
h:=RotatedHeight;
if (w=0) or (h=0) then Exit;

if Picture.Graphic is TIcon then ReadHandle(TIcon(Picture.Graphic).Handle);
mw:=Picture.Width;
mh:=Picture.Height;

Rotated:=GetMemory(w*h*4);
FillMemory(Rotated,w*h*4,0);
d:=Rotated;

si:=Sin(RR);
co:=Cos(RR);
mw2:=mw/2;
mh2:=mh/2;
w2:=w/2;                  
h2:=h/2;

for y:=h-1 downto 0 do
  for x:=0 to w-1 do
    begin
    c:=(x-w2)*co+(h2-y)*si+mw2;
    u:=(x-w2)*si-(h2-y)*co+mh2;
    if (c<=-1) or (u<=-1) or (c>=mw) or (u>=mh) then
      begin
      Inc(d);
      Continue;
      end;
    tc:=Tru(c);
    tu:=Tru(u);
    if (c<0) or (u<0) then l1:=BlendCol else l1:=GetPixel(Data,tc,tu,mw,mh);
    if (tc=mw-1) or (u<0) then l2:=BlendCol else l2:=GetPixel(Data,tc+1,tu,mw,mh);
    if (c<0) or (tu=mh-1) then l3:=BlendCol else l3:=GetPixel(Data,tc,tu+1,mw,mh);
    if (tc=mw-1) or (tu=mh-1) then l4:=BlendCol else l4:=GetPixel(Data,tc+1,tu+1,mw,mh);
    c:=c-tc;
    u:=u-tu;
    a1:=(1-c)*(1-u);
    a2:=c*(1-u);
    a3:=(1-c)*u;
    a4:=c*u;
    d^:=MakeColor;
    Inc(d);
    end;

MakeStretched;
end;

procedure TRotatableImage.MakeData;
var
  r,b:{$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap;
  x,y,w,h:Word;
  t,n:PByteArray;
  d:PCardinal;
begin
FreeMemory(Data);
Data:=nil;
FreeMemory(Rotated);
Rotated:=nil;
FreeMemory(Stretched);
Stretched:=nil;

if Picture.Graphic=nil then Exit;
if (RotatedWidth=0) or (RotatedHeight=0) then Exit;

if Picture.Graphic is TIcon then ReadHandle(TIcon(Picture.Graphic).Handle);
w:=Picture.Width;
h:=Picture.Height;

Data:=GetMemory(w*h*4);
d:=Data;

r:={$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap.Create;
r.Canvas.Pen.Color:=clRed;
r.Canvas.Brush.Color:=clRed;
r.Width:=w;
r.Height:=h;
r.Canvas.FillRect(r.Canvas.ClipRect);
r.Canvas.Draw(0,0,Picture.Graphic);
r.PixelFormat:=pf24bit;

b:={$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap.Create;
b.Canvas.Pen.Color:=clBlue;
b.Canvas.Brush.Color:=clBlue;
b.Width:=w;
b.Height:=h;
b.Canvas.FillRect(r.Canvas.ClipRect);
b.Canvas.Draw(0,0,Picture.Graphic);
b.PixelFormat:=pf24bit;

for y:=h-1 downto 0 do
  begin
  t:=r.ScanLine[y];
  n:=b.ScanLine[y];
  for x:=0 to w-1 do
    begin
    d^:=FindColor(t^[x*3+2],t^[x*3+1],t^[x*3],n^[x*3+2],n^[x*3+1],n^[x*3]);
    Inc(d);
    end;
  end;

r.Free;
b.Free;

MakeRotated;
MakeCustomMapping;
end;

function TRotatableImage.FindColor(RR,RG,RB,BR,BG,BB:Byte):Cardinal;
var r,g,b,a:Byte;
begin
if RGB(RR,RG,RB)=RGB(BR,BG,BB) then
  begin
  a:=255;
  r:=RR;
  g:=RG;
  b:=RB;
  end
else if (RR=255) and (BB=255) and (RG+RB+BR+BG=0) then
  begin
  Result:=BlendCol;
  Exit;
  end
else
  begin
  a:=BR-RR+255;
  r:=BR*255 div a;
  g:=BG*255 div a;
  b:=RB*255 div a;
  end;
Result:=r or (g shl 8) or (b shl 16) or (a shl 24);
end;

procedure TRotatableImage.ReadHandle(H:HIcon);
begin
//Forse icon to get the correct size
end;

procedure TRotatableImage.Paint;
var
  x,y,k,j,w,h:Integer;
  Save:Boolean;
  a,r:Real;
  b:{$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap;
  p:PByteArray;
  c:Cardinal;
  t:PByte;
  d:PCardinal;
begin
if csDesigning in ComponentState then
	with inherited Canvas do
	  begin
	  Pen.Style:=psDash;
	  Brush.Style:=bsClear;
	  Rectangle(0,0,Width,Height);
	  end;

if not (Visible and Assigned(Parent)) then Exit;
if Picture.Graphic=nil then Exit;
w:=RotatedWidth;
h:=RotatedHeight;
if (w=0) or (h=0) then Exit;

if Center then
  begin
  x:=(ClientWidth-w) div 2;
  y:=(ClientHeight-h) div 2;
  end
else
  begin
  x:=0;
  y:=0;
  end;

Save:=fDrawing;
fDrawing:=True;
try
  if fCM then
    begin
    if CustomMap=nil then
      begin
      fDrawing:=Save;
      Exit;
      end;
    b:={$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap.Create;
    b.Width:=ClientWidth;
    b.Height:=ClientHeight;
    b.Canvas.CopyRect(b.Canvas.ClipRect,inherited Canvas,ClientRect);
    b.PixelFormat:=pf32bit;
    t:=b.ScanLine[ClientHeight-1];
    d:=CustomMap;
    for j:=ClientHeight-1 downto 0 do
      for k:=0 to ClientWidth-1 do
        begin
        a:=Byte(d^ shr 24)/255;
        if a=0 then
          begin
          Inc(d);
          Inc(t,4);
          end
        else if a=1 then
          begin
          t^:=Byte(d^ shr 16);
          Inc(t);
          t^:=Byte(d^ shr 8);
          Inc(t);
          t^:=Byte(d^);
          Inc(t,2);
          Inc(d);
          end
        else
          begin
          t^:=Round((1-a)*t^+a*Byte(d^ shr 16));
          Inc(t);
          t^:=Round((1-a)*t^+a*Byte(d^ shr 8));
          Inc(t);
          t^:=Round((1-a)*t^+a*Byte(d^));
          Inc(t,2);
          Inc(d);
          end;
        end;
    inherited Canvas.Draw(0,0,b);
    b.Free;
    end
  else if Stretch or (Proportional and ((w>ClientWidth) or (h>ClientHeight))) then
    begin
    //draw Stretched
    if Stretched=nil then
      begin
      fDrawing:=Save;
      Exit;
      end;
    w:=StretchedWidth;
    h:=StretchedHeight;
    if Center then
      begin
      x:=(ClientWidth-w) div 2;
      y:=(ClientHeight-h) div 2;
      end;
    b:={$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap.Create;
    b.Width:=ClientWidth;
    b.Height:=ClientHeight;
    b.Canvas.CopyRect(b.Canvas.ClipRect,inherited Canvas,ClientRect);
    b.PixelFormat:=pf24bit;
    for j:=0 to ClientHeight+Abs(y)-1 do
      begin
      if (j+y<0) or (j+y>=ClientHeight) then Continue;
      if j>=h then Continue;
      p:=b.ScanLine[j+y];
      for k:=0 to ClientWidth+Abs(x)-1 do
        begin
        if (k+x<0) or (k+x>=ClientWidth) then Continue;
        if k>=w then Continue;
        c:=GetPixel(Stretched,k,j,w,h);
        a:=Byte(c shr 24)/255;
        if a=0 then Continue;
        if a=1 then
          begin
          p^[(k+x)*3]:=Byte(c shr 16);
          p^[(k+x)*3+1]:=Byte(c shr 8);
          p^[(k+x)*3+2]:=Byte(c);
          Continue;
          end;
        r:=p^[(k+x)*3]+(Byte(c shr 16)-p^[(k+x)*3])*a;
        p^[(k+x)*3]:=Round(r);
        r:=p^[(k+x)*3+1]+(Byte(c shr 8)-p^[(k+x)*3+1])*a;
        p^[(k+x)*3+1]:=Round(r);
        r:=p^[(k+x)*3+2]+(Byte(c)-p^[(k+x)*3+2])*a;
        p^[(k+x)*3+2]:=Round(r);
        end;
      end;
    inherited Canvas.Draw(0,0,b);
    b.Free;
    end
  else if RA<>0 then
    begin
    //draw Rotated
    if Rotated=nil then
      begin
      fDrawing:=Save;
      Exit;
      end;
    b:={$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap.Create;
    b.Width:=ClientWidth;
    b.Height:=ClientHeight;
    b.Canvas.CopyRect(b.Canvas.ClipRect,inherited Canvas,ClientRect);
    b.PixelFormat:=pf24bit;
    for j:=0 to ClientHeight+Abs(y)-1 do
      begin
      if (j+y<0) or (j+y>=ClientHeight) then Continue;
      if j>=h then Continue;
      p:=b.ScanLine[j+y];
      for k:=0 to ClientWidth+Abs(x)-1 do
        begin
        if (k+x<0) or (k+x>=ClientWidth) then Continue;
        if k>=w then Continue;
        c:=GetPixel(Rotated,k,j,w,h);
        a:=Byte(c shr 24)/255;
        if a=0 then Continue;
        if a=1 then
          begin
          p^[(k+x)*3]:=Byte(c shr 16);
          p^[(k+x)*3+1]:=Byte(c shr 8);
          p^[(k+x)*3+2]:=Byte(c);
          Continue;
          end;
        r:=p^[(k+x)*3]+(Byte(c shr 16)-p^[(k+x)*3])*a;
        p^[(k+x)*3]:=Round(r);
        r:=p^[(k+x)*3+1]+(Byte(c shr 8)-p^[(k+x)*3+1])*a;
        p^[(k+x)*3+1]:=Round(r);
        r:=p^[(k+x)*3+2]+(Byte(c)-p^[(k+x)*3+2])*a;
        p^[(k+x)*3+2]:=Round(r);
        end;
      end;
    inherited Canvas.Draw(0,0,b);
    b.Free;
    end
  else
    begin
    //draw Data
    inherited Canvas.Draw(x,y,Picture.Graphic);
    end;
finally
  fDrawing:=Save;
end;
end;

function TRotatableImage.RotatedWidth:Word;
var r:Single;
begin
if Picture.Graphic is TIcon then ReadHandle(TIcon(Picture.Graphic).Handle);
r:=Picture.Width*Abs(Cos(RR))+Picture.Height*Abs(Sin(RR));
Result:=Trunc(r);
if r-Result<>0 then Inc(Result);
end;

function TRotatableImage.RotatedHeight:Word;
var r:Single;
begin
if Picture.Graphic is TIcon then ReadHandle(TIcon(Picture.Graphic).Handle);
r:=Picture.Width*Abs(Sin(RR))+Picture.Height*Abs(Cos(RR));
Result:=Trunc(r);
if r-Result<>0 then Inc(Result);
end;

function TRotatableImage.StretchedWidth:Word;
var
  rw,rh:Word;
  cw,ch:Single;
begin
if Stretch then
  begin
  Result:=ClientWidth;
  Exit;
  end;
rw:=RotatedWidth;
rh:=RotatedHeight;
Result:=rw;
if (Proportional) and (rw>ClientWidth) then
  begin
  Result:=ClientWidth;
  cw:=ClientWidth/rw;
  end
else
  cw:=1;
if (Proportional) and (rh>ClientHeight) then
  ch:=ClientHeight/rh
else
  ch:=1;
if cw>ch then
  Result:=Round(rw*ch);
end;

function TRotatableImage.StretchedHeight:Word;
var
  rw,rh:Word;
  cw,ch:Single;
begin
if Stretch then
  begin
  Result:=ClientHeight;
  Exit;
  end;
rw:=RotatedWidth;
rh:=RotatedHeight;
Result:=rh;
if (Proportional) and (rw>ClientWidth) then
  cw:=ClientWidth/rw
else
  cw:=1;
if (Proportional) and (rh>ClientHeight) then
  begin
  Result:=ClientHeight;
  ch:=ClientHeight/rh;
  end
else
  ch:=1;
if cw<ch then
  Result:=Round(rh*cw);
end;

function TRotatableImage.GetPixel(Memory:Pointer;X,Y,Width,Height:Integer):Cardinal;
var p:PCardinal;
begin
p:=Memory;
Inc(p,(Height-Y-1)*Width+X);
Result:=p^;
end;

procedure TRotatableImage.Resize;
begin
MakeStretched;
MakeCustomMapping;
inherited Resize;
end;

procedure Register;
begin
  RegisterComponents('dihav', [TRotatableImage]);
end;

end.
